<?php
/**
 * fishuv login system
 *
 * @author fish <fish@archa.cz>
 * @package flogin
 * @version 0.5
 *
 */
/**
 * hlavni trida starajici se o prihlasovani
 * @package flogin
 *
 */
class flogin {

  /**
   * variable $dbh
   * @private objectr
   */
  private $dbh=NULL;
  /**
   * variable $sys_id
   * @private integer
   */
  private $sys_id=0;
  /**
   * variable is_logged
   * @private boolean je uzivatel prihlasenej ?
   */
  private $_logged=FALSE;
  /**
   * variable login_uid
   * @private integer ID uzivatele
   */
  private $_login_uid=0;
  /**
   * variable USER_GRANTS
   * @private array cache opravneni uzivatele
   */
  private $USER_GRANTS=array();
  /**
   * variable ACTIONS
   * @private array cache akci systemu
   */
  private $ACTIONS=array();

  /**
   * variable login_time
   * @public integer trvanlivost prihlaseni
   */
  public $login_time=28800; //8h
  /**
   * variable login_db
   * @public string schema s login tabulkama
   */
  public $login_db = NULL;
  /**
   * variable login_status
   * @public string vysledek procesu prihlaseni
   */
  public $login_status = "";
  /**
   * variable $OTPUsed
   * @public boolean pro prihlaseni bylo pouzito docasne heslo
   */
  public $OTPUsed = true;
  /**
   * variable $useSha1
   * @public boolean heslo je ulozene jako SHA1
   */
  public $useSha1 = true;
  /**
   * variable $disableIPCheck
   * @public boolean nekontrolovat IP uzivatele
   */
  public $disableIPCheck = FALSE;
  
  
  /**
   * variable $updateOldPasswords
   * @public boolean prepsat stara hesla do silnejsiho hashe
   */
  public $updateOldPasswords = true;

  

  // {{{ constructor
  /**
   * Constructor
   *
   * @param object $dbh volitelne, handle databazoveho pripojeni
   * @param integer $sys_id volitelne, id systemu
   * @return object flogin new flogin object
   */
  function __construct($dbh = NULL, $sys_id = NULL){
    if(isset($dbh)){
      $this->setDBH($dbh);
    }
    if(isset($sys_id)){
      $this->setSystemID($sys_id);
    }
  }
  // }}} constructor


  // {{{ setDBH
  /**
   * nastavi handle databaze
   *
   * @param object $dbh
   * @return void
   */
  function setDBH($dbh){
    $this->dbh=$dbh;
  }
  // }}} setDBH


  // {{{ setSystemID
  /**
   * nastavi ID systemu
   *
   * @param integer $sys_id
   * @return void
   */
  function setSystemID($sys_id){
    $this->sys_id=$sys_id;
  }
  // }}} setSystemID


  // {{{ logged
  /**
   * je uzivatel prihlasenej ?
   *
   * @return boolean
   */
  function logged(){
    return($this->_logged);
  }
  // }}} logged


  // {{{ login_uid
  /**
   * vrati ID prihlasenyho uzivatele
   *
   * @return integer
   */
  function login_uid(){
    return($this->_login_uid);
  }
  // }}} login_uid


  // {{{ checkLogin
  /**
   * autorizace uzivatele
   *
   * @param string $username
   * @param string $password
   * @return mixed
   */
  function checkLogin($username = NULL, $password = NULL){

    $this->_login_uid==0;
    $this->_is_logged=FALSE;
    $this->OTPUsed=FALSE;

    if(isset($username) && isset($password)){
      unset($_SESSION["flogin_uid"]);

      $query ="SELECT users.id_user,IF(user_actions.id_user IS NULL,MAX(NOT(group_actions.id_act IS NULL)),user_actions.permission) AS can_login, hash_pwd AS real_hash, users_otp.otp AS otp_hash FROM (";
      if(!empty($this->login_db)){
        $query.="`".$this->login_db."`.";
      }
      $query.="users, ";
      if(!empty($this->login_db)){
        $query.="`".$this->login_db."`.";
      }
      $query.="actions) ";
      $query.="LEFT JOIN ";
      if(!empty($this->login_db)){
        $query.="`".$this->login_db."`.";
      }

      $query.="user_actions ON (user_actions.id_user=users.id_user AND actions.id=user_actions.id_act) ";
      $query.="LEFT JOIN ";
      if(!empty($this->login_db)){
        $query.="`".$this->login_db."`.";
      }
      $query.="user_groups ON (user_groups.id_user=users.id_user) ";
      $query.="LEFT JOIN ";
      if(!empty($this->login_db)){
        $query.="`".$this->login_db."`.";
      }
      $query.="group_actions ON ((group_actions.id_grp=users.id_grp OR group_actions.id_grp=user_groups.id_grp) AND actions.id=group_actions.id_act)";
      $query.="LEFT JOIN ".(!empty($this->login_db)?"`".$this->login_db."`.":"")."users_otp ON users_otp.id_user=users.id_user AND users_otp.`expire`>NOW() ";
      $query.="WHERE login=".$this->dbh->quote($username)." ";

      $query.="AND actions.id_system= ".$this->dbh->quote($this->sys_id)." AND actions.varname REGEXP 'LOGIN$' ";
      $query.="GROUP BY users.id_user ";
      $query.="LIMIT 0,1 ";
      // echo $query._nl;
      
      $res=$this->dbh->query($query);
      if(PEAR::isError($res)){
        return(PEAR::throwError("Login error: ".__LINE__.": ".$res->getMessage()."; ".$res->getUserInfo()));
      } else {
        $row = $res->fetchRow();
        $userFound = !empty($row);
        $row["otp_used"] = 0;
        $hasValidPassword = false;
        $usedOldHashType = false;
        
        if ($this->checkPasswordHash($password, $row["real_hash"])) {
          $hasValidPassword = true;
        } elseif ($this->checkPasswordHash($password, $row["otp_hash"])) {
          $hasValidPassword = true;
          $row["otp_used"] = 1;
        } elseif ($this->checkPasswordHash($password, $row["real_hash"], false)) {
          // byl pouzit stary typ hashe
          $hasValidPassword = true;
          $usedOldHashType = true;
        }
        
        if($userFound && $hasValidPassword){
          $row["otp_used"]*=1;
          $row["id_user"]*=1;
          $row["can_login"]*=1;
          if($row["can_login"]){
            if($row["otp_used"]!=0){
              $this->OTPUsed=TRUE;
              $query ="DELETE FROM ".(!empty($this->login_db)?"`".$this->login_db."`.":"").".users_otp ";
              $query.="WHERE id=".($this->dbh->quote($row["otp_used"])*1)." ";
              $query.="LIMIT 1";
              $res=$this->dbh->query($query);
            }
            $query ="UPDATE ";
            if(!empty($this->login_db)){
              $query.="`".$this->login_db."`.";
            }
            $query.="users SET id_conn= ? ,lastact=NOW() ";
            if ($this->updateOldPasswords && $usedOldHashType && !$row["otp_used"]) {
              // pokud byl pouzit stary typ hashe pro bezne heslo a je povolena aktualizace,
              // provedu jeho zmenu na novy typ
              $query.= ",hash_pwd= ? ";
            }
            $query.="WHERE id_user= ? ";
            $sth=$this->dbh->prepare($query,NULL,MDB2_PREPARE_MANIP);
            if(!PEAR::isError($sth)){
              $queryParams = array();
              $queryParams[] = microtime().';'.$_SERVER["REMOTE_ADDR"].';'.$_SERVER["HTTP_USER_AGENT"];
              if ($this->updateOldPasswords && $usedOldHashType) {
                $queryParams[] = self::createHash($password);
              }
              $queryParams[] = $row["id_user"];
              if($qr=$sth->execute($queryParams)){
                if(PEAR::isError($qr)){
                  return(PEAR::throwError("Login error: ".__LINE__.": ".$qr->getMessage()."; ".$qr->getUserInfo()));
                } else {
                  if($qr==1){
                    $_SESSION["flogin_uid"]=$row["id_user"];
                    $_SESSION["flogin_stamp"]=time();
                    $_SESSION["flogin_ip"] = $_SERVER["REMOTE_ADDR"].(isset($_SERVER["HTTP_X_FORWARDED_FOR"])?";".$_SERVER["HTTP_X_FORWARDED_FOR"]:"");
                    $this->_login_uid=$row["id_user"];
                    $this->_logged=TRUE;
                    return(TRUE);
                  }
                }
              }
            } else {
              return(PEAR::throwError("Login error: ".__LINE__.": ".$sth->getMessage()."; ".$sth->getUserInfo()));
            }
          } else {
            $this->logout();
            $this->login_status=_("Váš login není aktivován");
            return(FALSE);
          }
        } else {
          $this->logout();
          $this->login_status=_("Neplatné přihlašovací jméno nebo heslo");
          return(FALSE);
        }
        $res->free();
      }
    } else {
      if(isset($_SESSION["flogin_uid"]) && $_SESSION["flogin_uid"]*1!=0){
        if(isset($_SESSION["flogin_ip"]) && !$this->disableIPCheck && $_SESSION["flogin_ip"]!=$_SERVER["REMOTE_ADDR"].(isset($_SERVER["HTTP_X_FORWARDED_FOR"])?";".$_SERVER["HTTP_X_FORWARDED_FOR"]:"")){
          $this->logout();
          $this->login_status=_("Byl jste odhlášen kvůli kolizi přihlášení");
          return(FALSE);
        } else {
          $last_stamp=0;
          if(isset($_SESSION["flogin_stamp"])){ $last_stamp=$_SESSION["flogin_stamp"]; }
          if(time()-$last_stamp>$this->login_time){
            $this->logout();
            if(time()-$last_stamp<$this->login_time+3600){
              //hlasku o odpojeni zobrazim jenom hodinu po vyprseni limitu
              $this->login_status=_("Byl jste odhlášen z důvodu dlouhé nečinnosti");
            }
            return(FALSE);
          } else {
#  kontrola jestli nema zakazany prihlaseni
            $query ="SELECT IF(user_actions.id_user IS NULL,MAX(NOT(group_actions.id_act IS NULL)),user_actions.permission) AS can_login FROM (";
            if(!empty($this->login_db)){
              $query.="`".$this->login_db."`.";
            }
            $query.="users, ";
            if(!empty($this->login_db)){
              $query.="`".$this->login_db."`.";
            }
            $query.="actions) ";
            $query.="LEFT JOIN ";
            if(!empty($this->login_db)){
              $query.="`".$this->login_db."`.";
            }
            $query.="user_actions ON (user_actions.id_user=users.id_user AND actions.id=user_actions.id_act) ";
            $query.="LEFT JOIN ";
            if(!empty($this->login_db)){
              $query.="`".$this->login_db."`.";
            }
            $query.="user_groups ON (user_groups.id_user=users.id_user) ";
            $query.="LEFT JOIN ";
            if(!empty($this->login_db)){
              $query.="`".$this->login_db."`.";
            }
            $query.="group_actions ON ((group_actions.id_grp=users.id_grp OR group_actions.id_grp=user_groups.id_grp) AND actions.id=group_actions.id_act)";
            $query.="WHERE users.id_user=".$this->dbh->quote($_SESSION["flogin_uid"]*1,'integer');
            $query.=" AND actions.id_system= ".$this->dbh->quote($this->sys_id)." AND actions.varname REGEXP 'LOGIN$' ";
            $query.="GROUP BY users.id_user ";
            $query.="LIMIT 0,1 ";
            $qr=$this->dbh->getOne($query);
            if(!PEAR::isError($qr)){
              if($qr==0){
                $this->logout();
                $this->login_status="Váš login není aktivován";
                return(FALSE);
              }
            } else {
              return(PEAR::throwError("Login error: ".__LINE__.": ".$qr->getMessage()."; ".$qr->getUserInfo()));
            }
#  update statusu
            $query ="UPDATE ";
            if(!empty($this->login_db)){
              $query.="`".$this->login_db."`.";
            }
            $query.="users SET id_conn= ? ,lastact=NOW() ";
            $query.="WHERE id_user= ? ";
            $sth=$this->dbh->prepare($query,NULL,MDB2_PREPARE_MANIP);
            if(!PEAR::isError($sth)){
              $qr=$sth->execute(array(microtime().';'.$_SERVER["REMOTE_ADDR"].';'.$_SERVER["HTTP_USER_AGENT"],$_SESSION["flogin_uid"]));
              if(PEAR::isError($qr)){
                return(PEAR::throwError("Login error: ".__LINE__.": ".$qr->getMessage()."; ".$qr->getUserInfo()));
              } else {
                if($qr==1){
                  $this->_logged=TRUE;
                  $this->_login_uid=$_SESSION["flogin_uid"];
                  $_SESSION["flogin_stamp"]=time();
                  $_SESSION["flogin_ip"]=$_SERVER["REMOTE_ADDR"].(isset($_SERVER["HTTP_X_FORWARDED_FOR"])?";".$_SERVER["HTTP_X_FORWARDED_FOR"]:"");
                  return(TRUE);
                }
              }
            } else {
              return(PEAR::throwError("Login error: ".__LINE__.": ".$sth->getMessage()."; ".$sth->getUserInfo()));
            }
          }
        }
      }
    }
    return(FALSE);
  }
  // }}} checkLogin


  // {{{ logout
  /**
   * odhlasi uzivatele
   *
   * @return void
   */
  function logout(){
    if(isset($_SESSION["flogin_uid"])){ unset($_SESSION["flogin_uid"]); }
    if(isset($_SESSION["flogin_stamp"])){ unset($_SESSION["flogin_stamp"]); }
    if(isset($_SESSION["flogin_ip"])){ unset($_SESSION["flogin_ip"]); }
    $this->_logged=FALSE;
    $this->_login_uid=0;
  }
  // }}} logout


  // {{{ canDoAction
  /**
   * overi jestli uzivatel muze provest danou akci
   *
   * @param integer $user_id ID uzivatele
   * @param string $action identifikator akce
   * @return boolean
   *
   */
  function canDoAction($user_id,$action){
    $user_id=(int)$user_id;
    $action=(string)$action;

    if($user_id==0 || $action==""){
      return(FALSE);
    } else {
      if(isset($this->USER_GRANTS[$user_id])){
        if(isset($this->USER_GRANTS[$user_id][$action])){
          return($this->USER_GRANTS[$user_id][$action]);
        } else {
          return(FALSE);
        }
      } else {
        $query ='SELECT gid FROM '.(!empty($this->login_db)?$this->dbh->quoteIdentifier($this->login_db).".":'').'users'.
          ' LEFT JOIN '.(!empty($this->login_db)?$this->dbh->quoteIdentifier($this->login_db).".":'').'user_groups ON (users.id_user=user_groups.id_user)'.
          ' LEFT JOIN '.(!empty($this->login_db)?$this->dbh->quoteIdentifier($this->login_db).".":'').'group_info ON (group_info.gid=users.id_grp OR user_groups.id_grp=group_info.gid)'.
          ' WHERE users.id_user='.$this->dbh->quote($user_id,'integer');
          // prettySQL($query);
        if (PEAR::isError($Groups=$this->dbh->getCol($query))) {
          return(PEAR::throwError("Login error: ".__LINE__.": ".$Groups->getMessage()."; ".$Groups->getUserInfo()));
        }
        $query ="SELECT actions.id_system,actions.varname,IF(user_actions.id_act IS NULL,MAX(NOT(group_actions.id_act IS NULL)),user_actions.permission) AS can_do FROM ";
        if(!empty($this->login_db)){
          $query.="`".$this->login_db."`.";
        }
        $query.="actions ";
        $query.="LEFT JOIN ";
        if(!empty($this->login_db)){
          $query.="`".$this->login_db."`.";
        }
        $query.="user_actions ON (user_actions.id_act=actions.id ";
        $query.="AND user_actions.id_user=".$this->dbh->quote($user_id).") ";
        $query.="LEFT JOIN ";
        if(!empty($this->login_db)){
          $query.="`".$this->login_db."`.";
        }
        $query.="group_actions ON (group_actions.id_grp IN ('".implode("','", $Groups)."') AND actions.id=group_actions.id_act)";
        $query.=" GROUP BY actions.id ";
        $res=$this->dbh->query($query);
        // prettySQL($query);
        if(!PEAR::isError($res)){
          if(isset($this->USER_GRANTS) && isset($this->USER_GRANTS[$user_id])){ unset($this->USER_GRANTS[$user_id]); }
          while($row=$res->fetchRow()){
            $this->USER_GRANTS[$user_id][$row["varname"]]=$row["can_do"]==1?TRUE:FALSE;
            if(!isset($this->USER_GRANTS[$user_id]["SYS_LOGIN"]) && preg_match("/LOGIN$/",$row["varname"]) && $row["id_system"]==$this->sys_id){
              $this->USER_GRANTS[$user_id]["SYS_LOGIN"]=$row["can_do"]==1?TRUE:FALSE;
            }
          }
          $res->free();
        } else {
          return(PEAR::throwError("Login error: ".__LINE__.": ".$res->getMessage()."; ".$res->getUserInfo()));
        }
        if(isset($this->USER_GRANTS[$user_id][$action])){
          return($this->USER_GRANTS[$user_id][$action]);
        }
      }
      return(FALSE);
    }
  }
  // }}} canDoAction

// {{{ setUserAct
  /**
   * nastaveni prava uzivatele
   *
   * @param object $dbh handle databazoveho pripojeni
   * @param integer $user_id id uzivatele
   * @param integer $sys_id id uzivatele
   * @param string $act nazev prava
   * @param boolean $val hodnota k nastaveni
   * @param string $login_db schema s login tabulkama
   * @return mixed TRUE | PEAR_Error
   */
  static function setUserAct(&$dbh,$user_id,$sys_id,$act,$val,$login_db=NULL){
    if($user_id>0){
      $query ="SELECT id FROM ";
      if(!empty($login_db)){ $query.="`".$login_db."`."; }
      $query.="actions ";
      $query.="WHERE ";
      if($act=="SYS_LOGIN"){
        $query.="varname REGEXP '_LOGIN$' ";
      } else{
        $query.="varname=".$dbh->quote($act)." ";
      }
      $query.="AND (id_system=0 OR id_system=".$dbh->quote($sys_id).") ";
      $id_act=0;
      $res=$dbh->query($query);
      if(!PEAR::isError($res)){
        if($row=$res->fetchRow()){
          $id_act=$row["id"];
        }
        $res->free();
      }
      if($id_act!=0){
        if($val!==null){
          $query ="INSERT IGNORE INTO ";
          if(!empty($login_db)){ $query.="`".$login_db."`."; }
          $query.="user_actions(id_user,id_act,permission) ";
          $query.="VALUES(".$dbh->quote($user_id).",".$dbh->quote($id_act).",".$dbh->quote($val).") ";
          if(!PEAR::isError($res=$dbh->query($query))){
            $query = 'UPDATE user_actions SET permission='.$dbh->quote($val).' WHERE id_user='.$dbh->quote($user_id).' AND id_act='.$dbh->quote($id_act);
          } else {
            return(PEAR::throwError("Login error: ".__LINE__.": ".$res->getMessage()."; ".$res->getUserInfo()));
          }
        } else {
          $query ="DELETE FROM ";
          if(!empty($login_db)){ $query.="`".$login_db."`."; }
          $query.="user_actions ";
          $query.="WHERE id_user=".$dbh->quote($user_id)." AND id_act=".$dbh->quote($id_act)." ";
        }
        if(!$res=$dbh->query($query)){
          return(PEAR::throwError("Login error: ".__LINE__.": ".$res->getMessage()."; ".$res->getUserInfo()));
        }
      } else {
        PEAR::throwError(__LINE__.": Neznámá akce ".$act);
      }
    }
  }
  // }}} setUserAct
  
  protected function checkPasswordHash($password, $hash, $strict = true)
  {
    if (strpos($hash, '$') === 0) {
      // ulozeny hash je novejsi pomoci crypt varianty
      return self::validateHash($password, $hash);
    } else {
      if ($strict) {
        // ve striktnim rezimu nepovolim kontrolu stareho hashe
        return false;
      } else {
        // hash je MD5 nebo SHA1, overim oboje
        return ($hash == sha1($password) || $hash == md5($password));
      }
    }
  }
  
  public static function validateHash($password, $hash)
  {
    return password_verify($password, $hash);
  }
  
  public static function createHash($password)
  {
    return password_hash($password, PASSWORD_DEFAULT);
  }

}
?>
